@@ -16,6 +16,10 @@
  * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
  */
 
+/*
+ * Admin console commands.
+ */
+
 #include "bouncer.h"
 
 #include <regex.h>
@@ -33,7 +37,7 @@
 #define CMD_NAME 1
 #define CMD_ARG 3
 #define SET_KEY 1
-#define SET_VAL 2
+#define SET_VAL 3
 
 typedef bool (*cmd_func_t)(PgSocket *admin, const char *arg);
 struct cmd_lookup {
@@ -43,15 +47,15 @@ struct cmd_lookup {
 
 /* CMD [arg]; */
 static const char cmd_normal_rx[] =
-"^" WS0 WORD "(" WS1 WORD ")?" WS0 ";" WS0 "$";
+"^" WS0 WORD "(" WS1 WORD ")?" WS0 "(;" WS0 ")?$";
 
 /* SET with simple value */
 static const char cmd_set_word_rx[] =
-"^" WS0 "set" WS1 WORD WS0 "=" WS0 WORD WS0 ";" WS0 "$";
+"^" WS0 "set" WS1 WORD WS0 "(=|to)" WS0 WORD WS0 "(;" WS0 ")?$";
 
 /* SET with quoted value */
 static const char cmd_set_str_rx[] =
-"^" WS0 "set" WS1 WORD WS0 "=" WS0 STRING WS0 ";" WS0 "$";
+"^" WS0 "set" WS1 WORD WS0 "(=|to)" WS0 STRING WS0 "(;" WS0 ")?$";
 
 /* compiled regexes */
 static regex_t rc_cmd;
@@ -114,28 +118,98 @@ static int count_db_active(PgDatabase *db)
 	return cnt;
 }
 
-void admin_flush(PgSocket *admin, PktBuf *buf, const char *desc)
+bool admin_flush(PgSocket *admin, PktBuf *buf, const char *desc)
 {
 	pktbuf_write_CommandComplete(buf, desc);
 	pktbuf_write_ReadyForQuery(buf);
-	pktbuf_send_queued(buf, admin);
+	return pktbuf_send_queued(buf, admin);
 }
 
 bool admin_ready(PgSocket *admin, const char *desc)
 {
 	PktBuf buf;
-	uint8 tmp[512];
+	uint8_t tmp[512];
 	pktbuf_static(&buf, tmp, sizeof(tmp));
 	pktbuf_write_CommandComplete(&buf, desc);
 	pktbuf_write_ReadyForQuery(&buf);
 	return pktbuf_send_immidiate(&buf, admin);
 }
 
+/*
+ * some silly clients start actively messing with server parameters
+ * without checking if thats necessary.  Fake some env for them.
+ */
+struct FakeParam {
+	const char *name;
+	const char *value;
+};
+
+static const struct FakeParam fake_param_list[] = {
+	{ "client_encoding", "UTF-8" },
+	{ "default_transaction_isolation", "read committed" },
+	{ "datestyle", "ISO" },
+	{ "timezone", "GMT" },
+	{ NULL },
+};
+
+/* fake result send, returns if handled */
+static bool fake_show(PgSocket *admin, const char *name)
+{
+	PktBuf *buf;
+	const struct FakeParam *p;
+	bool got = false;
+
+	for (p = fake_param_list; p->name; p++) {
+		if (strcasecmp(name, p->name) == 0) {
+			got = true;
+			break;
+		}
+	}
+
+	if (got) {
+		buf = pktbuf_dynamic(256);
+		if (buf) {
+			pktbuf_write_RowDescription(buf, "s", p->name);
+			pktbuf_write_DataRow(buf, "s", p->value);
+			admin_flush(admin, buf, "SHOW");
+		} else
+			admin_error(admin, "no mem");
+	}
+	return got;
+}
+
+static bool fake_set(PgSocket *admin, const char *key, const char *val)
+{
+	PktBuf *buf;
+	const struct FakeParam *p;
+	bool got = false;
+
+	for (p = fake_param_list; p->name; p++) {
+		if (strcasecmp(key, p->name) == 0) {
+			got = true;
+			break;
+		}
+	}
+
+	if (got) {
+		buf = pktbuf_dynamic(256);
+		if (buf) {
+			pktbuf_write_Notice(buf, "SET ignored");
+			admin_flush(admin, buf, "SET");
+		} else
+			admin_error(admin, "no mem");
+	}
+	return got;
+}
+
 /* Command: SET key = val; */
 static bool admin_set(PgSocket *admin, const char *key, const char *val)
 {
 	char tmp[512];
 
+	if (fake_set(admin, key, val))
+		return true;
+
 	if (admin->admin_user) {
 		if (set_config_param(bouncer_params, key, val, true, admin)) {
 			snprintf(tmp, sizeof(tmp), "SET %s=%s", key, val);
@@ -152,7 +226,7 @@ static bool send_one_fd(PgSocket *admin,
 			int fd, const char *task,
 			const char *user, const char *db,
 			const char *addr, int port,
-			uint64 ckey, int link,
+			uint64_t ckey, int link,
 			const char *client_enc,
 			const char *std_strings,
 			const char *datestyle,
@@ -162,8 +236,8 @@ static bool send_one_fd(PgSocket *admin,
 	struct cmsghdr *cmsg;
 	int res;
 	struct iovec iovec;
-	uint8 pktbuf[1024];
-	uint8 cntbuf[CMSG_SPACE(sizeof(int))];
+	uint8_t pktbuf[1024];
+	uint8_t cntbuf[CMSG_SPACE(sizeof(int))];
 
 	iovec.iov_base = pktbuf;
 	BUILD_DataRow(res, pktbuf, sizeof(pktbuf), "issssiqissss",
@@ -193,12 +267,12 @@ static bool send_one_fd(PgSocket *admin,
 	}
 
 	slog_debug(admin, "sending socket list: fd=%d, len=%d",
-		   fd, msg.msg_controllen);
+		   fd, (int)msg.msg_controllen);
 	res = safe_sendmsg(sbuf_socket(&admin->sbuf), &msg, 0);
 	if (res < 0) {
 		log_error("send_one_fd: sendmsg error: %s", strerror(errno));
 		return false;
-	} else if (res != iovec.iov_len) {
+	} else if ((size_t)res != iovec.iov_len) {
 		log_error("send_one_fd: partial sendmsg");
 		return false;
 	}
@@ -305,7 +379,7 @@ static bool admin_show_fds(PgSocket *admin, const char *arg)
 
 	statlist_for_each(item, &pool_list) {
 		pool = container_of(item, PgPool, head);
-		if (pool->admin)
+		if (pool->db->admin)
 			continue;
 		res = res && show_fds_from_list(admin, &pool->active_client_list);
 		res = res && show_fds_from_list(admin, &pool->waiting_client_list);
@@ -376,11 +450,11 @@ static bool admin_show_lists(PgSocket *admin, const char *arg)
 	SENDLIST("databases", statlist_count(&database_list));
 	SENDLIST("users", statlist_count(&user_list));
 	SENDLIST("pools", statlist_count(&pool_list));
-	SENDLIST("free_clients", statlist_count(&free_client_list));
-	SENDLIST("used_clients", get_active_client_count());
+	SENDLIST("free_clients", objcache_free_count(client_cache));
+	SENDLIST("used_clients", objcache_active_count(client_cache));
 	SENDLIST("login_clients", statlist_count(&login_client_list));
-	SENDLIST("free_servers", statlist_count(&free_server_list));
-	SENDLIST("used_servers", get_active_server_count());
+	SENDLIST("free_servers", objcache_free_count(server_cache));
+	SENDLIST("used_servers", objcache_active_count(server_cache));
 	admin_flush(admin, buf, "SHOW");
 	return true;
 }
@@ -419,22 +493,27 @@ static void socket_header(PktBuf *buf, bool debug)
 				    "pkt_avail", "send_avail");
 }
 
-static void adr2txt(const PgAddr *adr, char *dst, int dstlen)
+static void adr2txt(const PgAddr *adr, char *dst, unsigned dstlen)
 {
 	if (adr->is_unix) {
-		strlcpy(dst, "unix", dstlen);
+		safe_strcpy(dst, "unix", dstlen);
 	} else {
 		char *tmp = inet_ntoa(adr->ip_addr);
-		strlcpy(dst, tmp, dstlen);
+		safe_strcpy(dst, tmp, dstlen);
 	}
 }
 
 static void socket_row(PktBuf *buf, PgSocket *sk, const char *state, bool debug)
 {
-	int pkt_avail = sk->sbuf.recv_pos - sk->sbuf.pkt_pos;
-	int send_avail = sk->sbuf.recv_pos - sk->sbuf.send_pos;
+	int pkt_avail = 0, send_avail = 0;
 	char ptrbuf[128], linkbuf[128];
 	char l_addr[32], r_addr[32];
+	IOBuf *io = sk->sbuf.io;
+
+	if (io) {
+		pkt_avail = iobuf_amount_parse(sk->sbuf.io);
+		send_avail = iobuf_amount_pending(sk->sbuf.io);
+	}
 
 	adr2txt(&sk->remote_addr, r_addr, sizeof(r_addr));
 	adr2txt(&sk->local_addr, l_addr, sizeof(l_addr));
@@ -454,11 +533,11 @@ static void socket_row(PktBuf *buf, PgSocket *sk, const char *state, bool debug)
 			     sk->connect_time,
 			     sk->request_time,
 			     ptrbuf, linkbuf,
-			     sk->sbuf.recv_pos,
-			     sk->sbuf.pkt_pos,
+			     io ? io->recv_pos : 0,
+			     io ? io->parse_pos : 0,
 			     sk->sbuf.pkt_remain,
-			     sk->sbuf.send_pos,
-			     sk->sbuf.send_remain,
+			     io ? io->done_pos : 0,
+			     0,
 			     pkt_avail, send_avail);
 }
 
@@ -554,6 +633,46 @@ static bool admin_show_sockets(PgSocket *admin, const char *arg)
 	return true;
 }
 
+static void show_active_socket_list(PktBuf *buf, StatList *list, const char *state)
+{
+	List *item;
+	statlist_for_each(item, list) {
+		PgSocket *sk = container_of(item, PgSocket, head);
+		if (!sbuf_is_empty(&sk->sbuf))
+			socket_row(buf, sk, state, true);
+	}
+}
+
+/* Command: SHOW ACTIVE_SOCKETS */
+static bool admin_show_active_sockets(PgSocket *admin, const char *arg)
+{
+	List *item;
+	PgPool *pool;
+	PktBuf *buf;
+
+	buf = pktbuf_dynamic(256);
+	if (!buf) {
+		admin_error(admin, "no mem");
+		return true;
+	}
+
+	socket_header(buf, true);
+	statlist_for_each(item, &pool_list) {
+		pool = container_of(item, PgPool, head);
+		show_active_socket_list(buf, &pool->active_client_list, "cl_active");
+		show_active_socket_list(buf, &pool->waiting_client_list, "cl_waiting");
+
+		show_active_socket_list(buf, &pool->active_server_list, "sv_active");
+		show_active_socket_list(buf, &pool->idle_server_list, "sv_idle");
+		show_active_socket_list(buf, &pool->used_server_list, "sv_used");
+		show_active_socket_list(buf, &pool->tested_server_list, "sv_tested");
+		show_active_socket_list(buf, &pool->new_server_list, "sv_login");
+	}
+	show_active_socket_list(buf, &login_client_list, "cl_login");
+	admin_flush(admin, buf, "SHOW");
+	return true;
+}
+
 /* Command: SHOW POOLS */
 static bool admin_show_pools(PgSocket *admin, const char *arg)
 {
@@ -594,6 +713,33 @@ static bool admin_show_pools(PgSocket *admin, const char *arg)
 	return true;
 }
 
+static void slab_stat_cb(void *arg, const char *slab_name,
+			 unsigned size, unsigned free,
+			 unsigned total)
+{
+	PktBuf *buf = arg;
+	unsigned alloc = total * size;
+	pktbuf_write_DataRow(buf, "siiii", slab_name,
+			     size, total - free, free, alloc);
+}
+
+/* Command: SHOW MEM */
+static bool admin_show_mem(PgSocket *admin, const char *arg)
+{
+	PktBuf *buf;
+
+	buf = pktbuf_dynamic(256);
+	if (!buf) {
+		admin_error(admin, "no mem");
+		return true;
+	}
+	pktbuf_write_RowDescription(buf, "siiii", "name",
+				    "size", "used", "free", "memtotal");
+	objcache_stats(slab_stat_cb, buf);
+	admin_flush(admin, buf, "SHOW");
+	return true;
+}
+
 /* Command: SHOW CONFIG */
 static bool admin_show_config(PgSocket *admin, const char *arg)
 {
@@ -644,8 +790,14 @@ static bool admin_cmd_shutdown(PgSocket *admin, const char *arg)
 	if (!admin->admin_user)
 		return admin_error(admin, "admin access needed");
 
+	/*
+	 * note: new pooler expects unix socket file gone when it gets
+	 * event from fd.  currently atexit() cleanup should be called
+	 * before closing open sockets.
+	 */
 	log_info("SHUTDOWN command issued");
 	exit(0);
+
 	return true;
 }
 
@@ -655,6 +807,12 @@ static void full_resume(void)
 	cf_pause_mode = P_NONE;
 	if (tmp_mode == P_SUSPEND)
 		resume_all();
+
+	/* avoid surprise later if cf_shutdown stays set */
+	if (cf_shutdown) {
+		log_info("canceling shutdown");
+		cf_shutdown = 0;
+	}
 }
 
 /* Command: RESUME */
@@ -702,6 +860,8 @@ static bool admin_cmd_suspend(PgSocket *admin, const char *arg)
 	admin->wait_for_response = 1;
 	suspend_pooler();
 
+	g_suspend_start = get_cached_time();
+
 	return true;
 }
 
@@ -738,7 +898,7 @@ static bool admin_cmd_pause(PgSocket *admin, const char *arg)
 
 /* extract substring from regex group */
 static void copy_arg(const char *src, regmatch_t *glist,
-		     int gnum, char *dst, int dstmax)
+		     int gnum, char *dst, unsigned dstmax)
 {
 	regmatch_t *g = &glist[gnum];
 	unsigned len = g->rm_eo - g->rm_so;
@@ -777,13 +937,14 @@ static bool admin_show_help(PgSocket *admin, const char *arg)
 	SEND_generic(res, admin, 'N',
 		"sssss",
 		"SNOTICE", "C00000", "MConsole usage",
-		"D\n\tSHOW [HELP|CONFIG|DATABASES|FDS"
-		"|POOLS|CLIENTS|SERVERS|SOCKETS|LISTS|VERSION]\n"
+		"D\n\tSHOW HELP|CONFIG|DATABASES"
+		"|POOLS|CLIENTS|SERVERS|VERSION\n"
+		"\tSHOW FDS|SOCKETS|ACTIVE_SOCKETS|LISTS|MEM\n"
 		"\tSET key = arg\n"
 		"\tRELOAD\n"
-		"\tPAUSE\n"
+		"\tPAUSE [<db>]\n"
 		"\tSUSPEND\n"
-		"\tRESUME\n"
+		"\tRESUME [<db>]\n"
 		"\tSHUTDOWN", "");
 	if (res)
 		res = admin_ready(admin, "SHOW");
@@ -806,6 +967,12 @@ static bool admin_show_stats(PgSocket *admin, const char *arg)
 	return admin_database_stats(admin, &pool_list);
 }
 
+static bool admin_show_totals(PgSocket *admin, const char *arg)
+{
+	return show_stat_totals(admin, &pool_list);
+}
+
+
 static struct cmd_lookup show_map [] = {
 	{"clients", admin_show_clients},
 	{"config", admin_show_config},
@@ -816,14 +983,19 @@ static struct cmd_lookup show_map [] = {
 	{"pools", admin_show_pools},
 	{"servers", admin_show_servers},
 	{"sockets", admin_show_sockets},
+	{"active_sockets", admin_show_active_sockets},
 	{"stats", admin_show_stats},
 	{"users", admin_show_users},
 	{"version", admin_show_version},
+	{"totals", admin_show_totals},
+	{"mem", admin_show_mem},
 	{NULL, NULL}
 };
 
 static bool admin_cmd_show(PgSocket *admin, const char *arg)
 {
+	if (fake_show(admin, arg))
+		return true;
 	return exec_cmd(show_map, admin, arg, NULL);
 }
 
@@ -962,20 +1134,20 @@ void admin_setup(void)
 	db->addr.port = cf_listen_port;
 	db->addr.is_unix = 1;
 	db->pool_size = 2;
-	force_user(db, "pgbouncer", "");
+	db->admin = 1;
+	if (!force_user(db, "pgbouncer", ""))
+		fatal("no mem on startup - cannot alloc pgbouncer user");
 
-	/* fake pool, tag the it as special */
+	/* fake pool */
 	pool = get_pool(db, db->forced_user);
 	if (!pool)
 		fatal("cannot create admin pool?");
-	pool->admin = 1;
 	admin_pool = pool;
 
 	/* fake user, with disabled psw */
 	user = add_user("pgbouncer", "");
 	if (!user)
 		fatal("cannot create admin user?");
-	create_auth_cache();
 
 	/* prepare welcome */
 	pktbuf_static(&msg, pool->welcome_msg, sizeof(pool->welcome_msg));
@@ -1010,30 +1182,37 @@ void admin_pause_done(void)
 {
 	List *item, *tmp;
 	PgSocket *admin;
+	bool res;
 
 	statlist_for_each_safe(item, &admin_pool->active_client_list, tmp) {
 		admin = container_of(item, PgSocket, head);
 		if (!admin->wait_for_response)
 			continue;
 
+		res = false;
 		switch (cf_pause_mode) {
 		case P_PAUSE:
-			admin_ready(admin, "PAUSE");
+			res = admin_ready(admin, "PAUSE");
 			break;
 		case P_SUSPEND:
-			admin_ready(admin, "SUSPEND");
+			res = admin_ready(admin, "SUSPEND");
 			break;
 		default:
 			if (count_paused_databases() > 0)
-				admin_ready(admin, "PAUSE");
+				res = admin_ready(admin, "PAUSE");
 			else
+				/* fixme */
 				fatal("admin_pause_done: bad state");
 		}
-		admin->wait_for_response = 0;
+
+		if (!res)
+			disconnect_client(admin, false, "dead admin");
+		else
+			admin->wait_for_response = 0;
 	}
 
 	if (statlist_empty(&admin_pool->active_client_list)
-			&& cf_pause_mode == P_SUSPEND)
+	    && cf_pause_mode == P_SUSPEND)
 	{
 		log_info("Admin disappeared when suspended, doing RESUME");
 		cf_pause_mode = P_NONE;
