@@ -36,14 +36,14 @@ static bool check_client_passwd(PgSocket *client, const char *passwd)
 	case AUTH_PLAIN:
 		return strcmp(user->passwd, passwd) == 0;
 	case AUTH_CRYPT:
-		correct = crypt(user->passwd, (char *)client->salt);
+		correct = crypt(user->passwd, (char *)client->tmp_login_salt);
 		return strcmp(correct, passwd) == 0;
 	case AUTH_MD5:
 		if (strlen(passwd) != MD5_PASSWD_LEN)
 			return false;
 		if (!isMD5(user->passwd))
 			pg_md5_encrypt(user->passwd, user->name, strlen(user->name), user->passwd);
-		pg_md5_encrypt(user->passwd + 3, client->salt, 4, md5);
+		pg_md5_encrypt(user->passwd + 3, (char *)client->tmp_login_salt, 4, md5);
 		return strcmp(md5, passwd) == 0;
 	}
 	return false;
@@ -113,7 +113,10 @@ static bool decide_startup_pool(PgSocket *client, PktHdr *pkt)
 			username = val;
 		else if (varcache_set(&client->vars, key, val))
 			slog_debug(client, "got var: %s=%s", key, val);
-		else {
+		else if (strlist_contains(cf_ignore_startup_params, key)) {
+			slog_debug(client, "ignoring startup parameter: %s=%s", key, val);
+		} else {
+			slog_warning(client, "unsupported startup parameter: %s=%s", key, val);
 			disconnect_client(client, true, "Unknown startup parameter");
 			return false;
 		}
@@ -157,24 +160,24 @@ static const char valid_crypt_salt[] =
 
 static bool send_client_authreq(PgSocket *client)
 {
-	uint8 saltlen = 0;
+	uint8_t saltlen = 0;
 	int res;
 	int auth = cf_auth_type;
-	uint8 randbuf[2];
+	uint8_t randbuf[2];
 
 	if (auth == AUTH_CRYPT) {
 		saltlen = 2;
 		get_random_bytes(randbuf, saltlen);
-		client->salt[0] = valid_crypt_salt[randbuf[0] & SALT_MASK];
-		client->salt[1] = valid_crypt_salt[randbuf[1] & SALT_MASK];
-		client->salt[2] = 0;
+		client->tmp_login_salt[0] = valid_crypt_salt[randbuf[0] & SALT_MASK];
+		client->tmp_login_salt[1] = valid_crypt_salt[randbuf[1] & SALT_MASK];
+		client->tmp_login_salt[2] = 0;
 	} else if (cf_auth_type == AUTH_MD5) {
 		saltlen = 4;
-		get_random_bytes((void*)client->salt, saltlen);
+		get_random_bytes((void*)client->tmp_login_salt, saltlen);
 	} else if (auth == AUTH_ANY)
 		auth = AUTH_TRUST;
 
-	SEND_generic(res, client, 'R', "ib", auth, client->salt, saltlen);
+	SEND_generic(res, client, 'R', "ib", auth, client->tmp_login_salt, saltlen);
 	return res;
 }
 
@@ -220,7 +223,7 @@ static bool handle_client_startup(PgSocket *client, PktHdr *pkt)
 		if (!decide_startup_pool(client, pkt))
 			return false;
 
-		if (client->pool->admin) {
+		if (client->pool->db->admin) {
 			if (!admin_pre_login(client))
 				return false;
 		}
@@ -253,7 +256,7 @@ static bool handle_client_startup(PgSocket *client, PktHdr *pkt)
 		break;
 	case PKT_CANCEL:
 		if (mbuf_avail(&pkt->data) == BACKENDKEY_LEN) {
-			const uint8 *key = mbuf_get_bytes(&pkt->data, BACKENDKEY_LEN);
+			const uint8_t *key = mbuf_get_bytes(&pkt->data, BACKENDKEY_LEN);
 			memcpy(client->cancel_key, key, BACKENDKEY_LEN);
 			accept_cancel_request(client);
 		} else
@@ -304,7 +307,7 @@ static bool handle_client_work(PgSocket *client, PktHdr *pkt)
 			client->query_start = get_cached_time();
 		}
 
-		if (client->pool->admin)
+		if (client->pool->db->admin)
 			return admin_handle_client(client, pkt);
 
 		/* aquire server */
@@ -333,10 +336,10 @@ static bool handle_client_work(PgSocket *client, PktHdr *pkt)
 }
 
 /* callback from SBuf */
-bool client_proto(SBuf *sbuf, SBufEvent evtype, MBuf *data, void *arg)
+bool client_proto(SBuf *sbuf, SBufEvent evtype, MBuf *data)
 {
 	bool res = false;
-	PgSocket *client = arg;
+	PgSocket *client = container_of(sbuf, PgSocket, sbuf);
 	PktHdr pkt;
 
 
@@ -344,11 +347,9 @@ bool client_proto(SBuf *sbuf, SBufEvent evtype, MBuf *data, void *arg)
 	Assert(client->sbuf.sock);
 	Assert(client->state != CL_FREE);
 
-	if (client->state == CL_JUSTFREE) {
-		/* SBuf should catch the case */
-		slog_warning(client, "state=CL_JUSTFREE, should not happen");
+	/* may happen if close failed */
+	if (client->state == CL_JUSTFREE)
 		return false;
-	}
 
 	switch (evtype) {
 	case SBUF_EV_CONNECT_OK:
@@ -392,6 +393,9 @@ bool client_proto(SBuf *sbuf, SBufEvent evtype, MBuf *data, void *arg)
 	case SBUF_EV_FLUSH:
 		/* client is not interested in it */
 		break;
+	case SBUF_EV_PKT_CALLBACK:
+		/* unused ATM */
+		break;
 	}
 	return res;
 }
