@@ -38,28 +38,47 @@ static void close_server_list(StatList *sk_list, const char *reason)
 	}
 }
 
-bool suspend_socket(PgSocket *sk)
+static void close_client_list(StatList *sk_list, const char *reason)
 {
-	if (!sk->suspended) {
-		if (sbuf_is_empty(&sk->sbuf)) {
-			sbuf_pause(&sk->sbuf);
-			sk->suspended = 1;
-		} else
-			return false;
+	List *item, *tmp;
+	PgSocket *client;
+
+	statlist_for_each_safe(item, sk_list, tmp) {
+		client = container_of(item, PgSocket, head);
+		disconnect_client(client, true, reason);
 	}
+}
+
+bool suspend_socket(PgSocket *sk, bool force_suspend)
+{
+	if (sk->suspended)
+		return true;
+
+	if (sbuf_is_empty(&sk->sbuf)) {
+		if (sbuf_pause(&sk->sbuf))
+			sk->suspended = 1;
+	}
+
+	if (sk->suspended || !force_suspend)
+		return sk->suspended;
+
+	if (is_server_socket(sk))
+		disconnect_server(sk, true, "suspend_timeout");
+	else
+		disconnect_client(sk, true, "suspend_timeout");
 	return true;
 }
 
 /* suspend all sockets in socket list */
-static int suspend_socket_list(StatList *list)
+static int suspend_socket_list(StatList *list, bool force_suspend)
 {
-	List *item;
+	List *item, *tmp;
 	PgSocket *sk;
 	int active = 0;
 
-	statlist_for_each(item, list) {
+	statlist_for_each_safe(item, list, tmp) {
 		sk = container_of(item, PgSocket, head);
-		if (!suspend_socket(sk))
+		if (!suspend_socket(sk, force_suspend))
 			active++;
 	}
 	return active;
@@ -88,7 +107,7 @@ static void resume_sockets(void)
 
 	statlist_for_each(item, &pool_list) {
 		pool = container_of(item, PgPool, head);
-		if (pool->admin)
+		if (pool->db->admin)
 			continue;
 		resume_socket_list(&pool->active_client_list);
 		resume_socket_list(&pool->active_server_list);
@@ -188,7 +207,7 @@ static int per_loop_pause(PgPool *pool)
 {
 	int active = 0;
 
-	if (pool->admin)
+	if (pool->db->admin)
 		return 0;
 
 	close_server_list(&pool->idle_server_list, "pause mode");
@@ -204,23 +223,23 @@ static int per_loop_pause(PgPool *pool)
 /*
  * suspend active clients and servers
  */
-static int per_loop_suspend(PgPool *pool)
+static int per_loop_suspend(PgPool *pool, bool force_suspend)
 {
 	int active = 0;
 
-	if (pool->admin)
+	if (pool->db->admin)
 		return 0;
 
-	active += suspend_socket_list(&pool->active_client_list);
+	active += suspend_socket_list(&pool->active_client_list, force_suspend);
 
-	if (!statlist_empty(&pool->waiting_client_list)) {
-		active += statlist_count(&pool->waiting_client_list);
+	/* this list is unsuspendable, but still need force_suspend and counting */
+	active += suspend_socket_list(&pool->waiting_client_list, force_suspend);
+	if (active)
 		per_loop_activate(pool);
-	}
 
 	if (!active) {
-		active += suspend_socket_list(&pool->active_server_list);
-		active += suspend_socket_list(&pool->idle_server_list);
+		active += suspend_socket_list(&pool->active_server_list, force_suspend);
+		active += suspend_socket_list(&pool->idle_server_list, force_suspend);
 
 		/* as all clients are done, no need for them */
 		close_server_list(&pool->tested_server_list, "close unsafe file descriptors on suspend");
@@ -239,14 +258,17 @@ void per_loop_maint(void)
 	PgPool *pool;
 	int active = 0;
 	int partial_pause = 0;
+	bool force_suspend = false;
 
-	/* don't touch anything if takeover is in progress */
-	if (cf_reboot)
-		return;
+	if (cf_pause_mode == P_SUSPEND && cf_suspend_timeout > 0) {
+		usec_t stime = get_cached_time() - g_suspend_start;
+		if (stime >= cf_suspend_timeout)
+			force_suspend = true;
+	}
 
 	statlist_for_each(item, &pool_list) {
 		pool = container_of(item, PgPool, head);
-		if (pool->admin)
+		if (pool->db->admin)
 			continue;
 		switch (cf_pause_mode) {
 		case P_NONE:
@@ -260,14 +282,17 @@ void per_loop_maint(void)
 			active += per_loop_pause(pool);
 			break;
 		case P_SUSPEND:
-			active += per_loop_suspend(pool);
+			active += per_loop_suspend(pool, force_suspend);
 			break;
 		}
 	}
 
 	switch (cf_pause_mode) {
 	case P_SUSPEND:
-		active += statlist_count(&login_client_list);
+		if (force_suspend) {
+			close_client_list(&login_client_list, "suspend_timeout");
+		} else
+			active += statlist_count(&login_client_list);
 	case P_PAUSE:
 		if (!active)
 			admin_pause_done();
@@ -295,7 +320,7 @@ static void pool_client_maint(PgPool *pool)
 			if (client->link)
 				continue;
 			if (now - client->request_time > cf_client_idle_timeout)
-				disconnect_client(client, true, "idle_timeout");
+				disconnect_client(client, true, "client_idle_timeout");
 		}
 	}
 
@@ -328,7 +353,7 @@ static void check_unused_servers(PgPool *pool, StatList *slist, bool idle_test)
 	 * must be separated.  This avoids the need to re-launch lot
 	 * of connections together.
 	 */
-	if (cf_server_lifetime > 0 && pool->db->pool_size > 0)
+	if (pool->db->pool_size > 0)
 		lifetime_kill_gap = cf_server_lifetime / pool->db->pool_size;
 
 	/* disconnect idle servers if needed */
@@ -346,7 +371,7 @@ static void check_unused_servers(PgPool *pool, StatList *slist, bool idle_test)
 			disconnect_server(server, true, "SV_USED server got dirty");
 		} else if (cf_server_idle_timeout > 0 && idle > cf_server_idle_timeout) {
 			disconnect_server(server, true, "server idle timeout");
-		} else if (cf_server_lifetime > 0 && age > cf_server_lifetime) {
+		} else if (age >= cf_server_lifetime) {
 			if (pool->last_lifetime_disconnect + lifetime_kill_gap <= now) {
 				disconnect_server(server, true, "server lifetime over");
 				pool->last_lifetime_disconnect = now;
@@ -457,13 +482,9 @@ static void do_full_maint(int sock, short flags, void *arg)
 	List *item;
 	PgPool *pool;
 
-	/* don't touch anything if takeover is in progress */
-	if (cf_reboot)
-		goto skip;
-
 	statlist_for_each(item, &pool_list) {
 		pool = container_of(item, PgPool, head);
-		if (pool->admin)
+		if (pool->db->admin)
 			continue;
 		pool_server_maint(pool);
 		pool_client_maint(pool);
@@ -479,8 +500,7 @@ static void do_full_maint(int sock, short flags, void *arg)
 	if (cf_auth_type >= AUTH_TRUST)
 		loader_users_check();
 
-skip:
-	evtimer_add(&full_maint_ev, &full_maint_period);
+	safe_evtimer_add(&full_maint_ev, &full_maint_period);
 }
 
 /* first-time initializtion */
@@ -488,19 +508,60 @@ void janitor_setup(void)
 {
 	/* launch maintenance */
 	evtimer_set(&full_maint_ev, do_full_maint, NULL);
-	evtimer_add(&full_maint_ev, &full_maint_period);
+	safe_evtimer_add(&full_maint_ev, &full_maint_period);
+}
+
+static void kill_pool(PgPool *pool)
+{
+	const char *reason = "database removed";
+
+	close_client_list(&pool->active_client_list, reason);
+	close_client_list(&pool->waiting_client_list, reason);
+	close_client_list(&pool->cancel_req_list, reason);
+
+	close_server_list(&pool->active_server_list, reason);
+	close_server_list(&pool->idle_server_list, reason);
+	close_server_list(&pool->used_server_list, reason);
+	close_server_list(&pool->tested_server_list, reason);
+	close_server_list(&pool->new_server_list, reason);
+
+	list_del(&pool->map_head);
+	statlist_remove(&pool->head, &pool_list);
+	obj_free(pool_cache, pool);
+}
+
+static void kill_database(PgDatabase *db)
+{
+	PgPool *pool;
+	List *item, *tmp;
+
+	log_warning("dropping database '%s' as it does not exist anymore", db->name);
+
+	statlist_for_each_safe(item, &pool_list, tmp) {
+		pool = container_of(item, PgPool, head);
+		if (pool->db == db)
+			kill_pool(pool);
+	}
+	if (db->forced_user)
+		obj_free(user_cache, db->forced_user);
+	if (db->connect_query)
+		free((void *)db->connect_query);
+	statlist_remove(&db->head, &database_list);
+	obj_free(db_cache, db);
 }
 
 /* as [pgbouncer] section can be loaded after databases,
    there's need for review */
 void config_postprocess(void)
 {
-	List *item;
+	List *item, *tmp;
 	PgDatabase *db;
 
-	statlist_for_each(item, &database_list) {
+	statlist_for_each_safe(item, &database_list, tmp) {
 		db = container_of(item, PgDatabase, head);
-		if (db->pool_size < 0)
+		if (db->db_dead)
+			kill_database(db);
+		else if (db->pool_size < 0)
 			db->pool_size = cf_default_pool_size;
 	}
 }
