@@ -31,13 +31,21 @@ static bool set_mode(ConfElem *elem, const char *val, PgSocket *console);
 static const char *get_mode(ConfElem *elem);
 static bool set_auth(ConfElem *elem, const char *val, PgSocket *console);
 static const char *get_auth(ConfElem *elem);
+static bool set_defer_accept(ConfElem *elem, const char *val, PgSocket *console);
 
 static const char *usage_str =
-"usage: pgbouncer [-d] [-R] [-q] [-v] [-h|-V] config.ini\n";
+"Usage: %s [OPTION]... config.ini\n"
+"  -d            Run in background (as a daemon)\n"
+"  -R            Do a online restart\n"
+"  -q            Run quietly\n"
+"  -v            Increase verbosity\n"
+"  -u <username> Assume identity of <username>\n"
+"  -V            Show version\n"
+"  -h            Show this help screen and exit\n";
 
-static void usage(int err)
+static void usage(int err, char *exe)
 {
-	printf(usage_str);
+	printf(usage_str, basename(exe));
 	exit(err);
 }
 
@@ -51,6 +59,9 @@ int cf_daemon = 0;
 int cf_pause_mode = P_NONE;
 int cf_shutdown = 0;
 int cf_reboot = 0;
+int cf_syslog = 0;
+static char *cf_username = "";
+char *cf_syslog_facility = "daemon";
 static char *cf_config_file;
 
 char *cf_listen_addr = NULL;
@@ -62,8 +73,8 @@ int cf_pool_mode = POOL_SESSION;
 /* sbuf config */
 int cf_sbuf_len = 2048;
 int cf_tcp_socket_buffer = 0;
-#ifdef TCP_DEFER_ACCEPT
-int cf_tcp_defer_accept = 45;
+#if defined(TCP_DEFER_ACCEPT) || defined(SO_ACCEPTFILTER)
+int cf_tcp_defer_accept = 1;
 #else
 int cf_tcp_defer_accept = 0;
 #endif
@@ -83,6 +94,8 @@ char *cf_server_check_query = "select 1";
 usec_t cf_server_check_delay = 30 * USEC;
 int cf_server_round_robin = 0;
 
+char *cf_ignore_startup_params = "";
+
 usec_t cf_server_lifetime = 60*60*USEC;
 usec_t cf_server_idle_timeout = 10*60*USEC;
 usec_t cf_server_connect_timeout = 15*USEC;
@@ -90,10 +103,13 @@ usec_t cf_server_login_retry = 15*USEC;
 usec_t cf_query_timeout = 0*USEC;
 usec_t cf_client_idle_timeout = 0*USEC;
 usec_t cf_client_login_timeout = 60*USEC;
+usec_t cf_suspend_timeout = 10*USEC;
+
+usec_t g_suspend_start = 0;
 
 char *cf_logfile = NULL;
 char *cf_pidfile = NULL;
-static char *cf_jobname = NULL;
+char *cf_jobname = "pgbouncer";
 
 char *cf_admin_users = "";
 char *cf_stats_users = "";
@@ -119,6 +135,9 @@ ConfElem bouncer_params[] = {
 {"pool_mode",		true, {get_mode, set_mode}},
 {"max_client_conn",	true, CF_INT, &cf_max_client_conn},
 {"default_pool_size",	true, CF_INT, &cf_default_pool_size},
+{"syslog",		true, CF_INT, &cf_syslog},
+{"syslog_facility",	true, CF_STR, &cf_syslog_facility},
+{"user",		false, CF_STR, &cf_username},
 
 {"server_reset_query",	true, CF_STR, &cf_server_reset_query},
 {"server_check_query",	true, CF_STR, &cf_server_check_query},
@@ -131,9 +150,11 @@ ConfElem bouncer_params[] = {
 {"server_connect_timeout",true, CF_TIME, &cf_server_connect_timeout},
 {"server_login_retry",	true, CF_TIME, &cf_server_login_retry},
 {"server_round_robin",	true, CF_INT, &cf_server_round_robin},
+{"suspend_timeout",	true, CF_TIME, &cf_suspend_timeout},
+{"ignore_startup_parameters", true, CF_STR, &cf_ignore_startup_params},
 
 {"pkt_buf",		false, CF_INT, &cf_sbuf_len},
-{"tcp_defer_accept",	false, CF_INT, &cf_tcp_defer_accept},
+{"tcp_defer_accept",	true, {cf_get_int, set_defer_accept}, &cf_tcp_defer_accept},
 {"tcp_socket_buffer",	true, CF_INT, &cf_tcp_socket_buffer},
 {"tcp_keepalive",	true, CF_INT, &cf_tcp_keepalive},
 {"tcp_keepcnt",		true, CF_INT, &cf_tcp_keepcnt},
@@ -215,18 +236,49 @@ static bool set_auth(ConfElem *elem, const char *val, PgSocket *console)
 	return true;
 }
 
+static bool set_defer_accept(ConfElem *elem, const char *val, PgSocket *console)
+{
+	bool ok;
+	int oldval = cf_tcp_defer_accept;
+	ok = cf_set_int(elem, val, console);
+	if (ok && !!oldval != !!cf_tcp_defer_accept)
+		pooler_tune_accept(cf_tcp_defer_accept);
+	return true;
+}
+
+static void set_dbs_dead(bool flag)
+{
+	List *item;
+	PgDatabase *db;
+
+	statlist_for_each(item, &database_list) {
+		db = container_of(item, PgDatabase, head);
+		if (db->admin)
+			continue;
+		db->db_dead = flag;
+	}
+}
+
 /* config loading, tries to be tolerant to errors */
 void load_config(bool reload)
 {
+	bool ok;
+
+	set_dbs_dead(true);
+
 	/* actual loading */
-	iniparser(cf_config_file, bouncer_config, reload);
+	ok = iniparser(cf_config_file, bouncer_config, reload);
+	if (ok) {
+		/* load users if needed */
+		if (cf_auth_type >= AUTH_TRUST)
+			load_auth_file(cf_auth_file);
 
-	/* load users if needed */
-	if (cf_auth_type >= AUTH_TRUST)
-		load_auth_file(cf_auth_file);
-
-	/* reset pool_size */
-	config_postprocess();
+		/* reset pool_size, kill dbs */
+		config_postprocess();
+	} else {
+		/* if ini file missing, dont kill anybody */
+		set_dbs_dead(false);
+	}
 
 	/* reopen logfile */
 	if (reload)
@@ -257,13 +309,15 @@ static void handle_sigint(int sock, short flags, void *arg)
 	log_info("Got SIGINT, shutting down");
 	if (cf_reboot)
 		fatal("Takeover was in progress, going down immediately");
+	if (cf_pause_mode == P_SUSPEND)
+		fatal("Suspend was in progress, going down immediately");
 	cf_pause_mode = P_PAUSE;
 	cf_shutdown = 1;
 }
 
 static void handle_sigusr1(int sock, short flags, void *arg)
 {
-	if (cf_pause_mode == 0) {
+	if (cf_pause_mode == P_NONE) {
 		log_info("Got SIGUSR1, pausing all activity");
 		cf_pause_mode = P_PAUSE;
 	} else {
@@ -277,15 +331,21 @@ static void handle_sigusr2(int sock, short flags, void *arg)
 	case P_SUSPEND:
 		log_info("Got SIGUSR2, continuing from SUSPEND");
 		resume_all();
-		cf_pause_mode = 0;
+		cf_pause_mode = P_NONE;
 		break;
 	case P_PAUSE:
 		log_info("Got SIGUSR2, continuing from PAUSE");
-		cf_pause_mode = 0;
+		cf_pause_mode = P_NONE;
 		break;
 	case P_NONE:
 		log_info("Got SIGUSR1, but not paused/suspended");
 	}
+
+	/* avoid surprise later if cf_shutdown stays set */
+	if (cf_shutdown) {
+		log_info("Canceling shutdown");
+		cf_shutdown = 0;
+	}
 }
 
 static void handle_sighup(int sock, short flags, void *arg)
@@ -307,16 +367,31 @@ static void signal_setup(void)
 		fatal_perror("sigprocmask");
 
 	/* install handlers */
+
 	signal_set(&ev_sigterm, SIGTERM, handle_sigterm, NULL);
-	signal_add(&ev_sigterm, NULL);
+	err = signal_add(&ev_sigterm, NULL);
+	if (err < 0)
+		fatal_perror("signal_add");
+
 	signal_set(&ev_sigint, SIGINT, handle_sigint, NULL);
-	signal_add(&ev_sigint, NULL);
+	err = signal_add(&ev_sigint, NULL);
+	if (err < 0)
+		fatal_perror("signal_add");
+
 	signal_set(&ev_sigusr1, SIGUSR1, handle_sigusr1, NULL);
-	signal_add(&ev_sigusr1, NULL);
+	err = signal_add(&ev_sigusr1, NULL);
+	if (err < 0)
+		fatal_perror("signal_add");
+
 	signal_set(&ev_sigusr2, SIGUSR2, handle_sigusr2, NULL);
-	signal_add(&ev_sigusr2, NULL);
+	err = signal_add(&ev_sigusr2, NULL);
+	if (err < 0)
+		fatal_perror("signal_add");
+
 	signal_set(&ev_sighup, SIGHUP, handle_sighup, NULL);
-	signal_add(&ev_sighup, NULL);
+	err = signal_add(&ev_sighup, NULL);
+	if (err < 0)
+		fatal_perror("signal_add");
 }
 
 /*
@@ -364,16 +439,8 @@ static void go_daemon(void)
 }
 
 /*
- * write pidfile.  if exists, quit with error.
+ * pidfile management.
  */
-static void check_pidfile(void)
-{
-	struct stat st;
-	if (!cf_pidfile)
-		return;
-	if (stat(cf_pidfile, &st) >= 0)
-		fatal("pidfile exists, another instance running?");
-}
 
 static void remove_pidfile(void)
 {
@@ -382,6 +449,53 @@ static void remove_pidfile(void)
 	unlink(cf_pidfile);
 }
 
+static void check_pidfile(void)
+{
+	char buf[128 + 1];
+	struct stat st;
+	pid_t pid = 0;
+	int fd, res;
+
+	if (!cf_pidfile)
+		return;
+
+	/* check if pidfile exists */
+	if (stat(cf_pidfile, &st) < 0) {
+		if (errno != ENOENT)
+			fatal_perror("stat");
+		return;
+	}
+
+	/* read old pid */
+	fd = open(cf_pidfile, O_RDONLY);
+	if (fd < 0)
+		goto locked_pidfile;
+	res = read(fd, buf, sizeof(buf) - 1);
+	close(fd);
+	if (res <= 0)
+		goto locked_pidfile;
+
+	/* parse pid */
+	buf[res] = 0;
+	pid = atol(buf);
+	if (pid <= 0)
+		goto locked_pidfile;
+
+	/* check if running */
+	if (kill(pid, 0) >= 0)
+		goto locked_pidfile;
+	if (errno != ESRCH)
+		goto locked_pidfile;
+
+	/* seems the pidfile is not in use */
+	log_info("Stale pidfile, removing");
+	remove_pidfile();
+	return;
+
+locked_pidfile:
+	fatal("pidfile exists, another instance running?");
+}
+
 static void write_pidfile(void)
 {
 	char buf[64];
@@ -410,39 +524,102 @@ static void write_pidfile(void)
 static void check_limits(void)
 {
 	struct rlimit lim;
-	int err = getrlimit(RLIMIT_NOFILE, &lim);
-	if (err < 0)
+	int total_users = statlist_count(&user_list);
+	int fd_count;
+	int err;
+	List *item;
+	PgDatabase *db;
+
+	log_noise("event: %d, SBuf: %d, PgSocket: %d, IOBuf: %d",
+		  (int)sizeof(struct event), (int)sizeof(SBuf),
+		  (int)sizeof(PgSocket), (int)IOBUF_SIZE);
+
+	/* load limits */
+	err = getrlimit(RLIMIT_NOFILE, &lim);
+	if (err < 0) {
 		log_error("could not get RLIMIT_NOFILE: %s", strerror(errno));
-	else
-		log_info("File descriptors limits: S:%d H:%d",
-			 (int)lim.rlim_cur, (int)lim.rlim_max);
+		return;
+	}
+
+	/* calculate theoretical max, +10 is just in case */
+	fd_count = cf_max_client_conn + 10;
+	statlist_for_each(item, &database_list) {
+		db = container_of(item, PgDatabase, head);
+		if (db->forced_user)
+			fd_count += db->pool_size;
+		else
+			fd_count += db->pool_size * total_users;
+	}
+
+	log_info("File descriptor limit: %d (H:%d), max_client_conn: %d, max fds possible: %d",
+		 (int)lim.rlim_cur, (int)lim.rlim_max, cf_max_client_conn, fd_count);
 }
 
-static void daemon_setup(void)
+static bool check_old_process_unix(void)
 {
-	if (!cf_reboot)
-		check_pidfile();
-	if (cf_daemon)
-		go_daemon();
-	if (!cf_reboot)
-		write_pidfile();
+	struct sockaddr_un sa_un;
+	socklen_t len = sizeof(sa_un);
+	int domain = AF_UNIX;
+	int res, fd;
+
+	if (!*cf_unix_socket_dir)
+		return false;
+
+	memset(&sa_un, 0, len);
+	sa_un.sun_family = domain;
+	snprintf(sa_un.sun_path, sizeof(sa_un.sun_path),
+		 "%s/.s.PGSQL.%d", cf_unix_socket_dir, cf_listen_port);
+
+	fd = socket(domain, SOCK_STREAM, 0);
+	if (fd < 0)
+		fatal_perror("cannot create socket");
+	res = safe_connect(fd, (struct sockaddr *)&sa_un, len);
+	safe_close(fd);
+	if (res < 0)
+		return false;
+	return true;
 }
 
 static void main_loop_once(void)
 {
+	int err;
+
 	reset_time_cache();
-	event_loop(EVLOOP_ONCE);
+
+	err = event_loop(EVLOOP_ONCE);
+	if (err < 0) {
+		if (errno != EINTR)
+			log_warning("event_loop failed: %s", strerror(errno));
+	}
 	per_loop_maint();
 	reuse_just_freed_objects();
+	rescue_timers();
+	per_loop_pooler_maint();
+}
+
+static void takeover_part1(void)
+{
+	/* use temporary libevent base */
+	void *evtmp = event_init();
+
+	if (!*cf_unix_socket_dir)
+		fatal("cannot reboot if unix dir not configured");
+
+	takeover_init();
+	while (cf_reboot)
+		main_loop_once();
+	event_base_free(evtmp);
 }
 
 /* boot everything */
 int main(int argc, char *argv[])
 {
 	int c;
+	bool did_takeover = false;
+	char *arg_username = NULL;
 
 	/* parse cmdline */
-	while ((c = getopt(argc, argv, "avhdVR")) != EOF) {
+	while ((c = getopt(argc, argv, "avhdVRu:")) != EOF) {
 		switch (c) {
 		case 'R':
 			cf_reboot = 1;
@@ -459,38 +636,72 @@ int main(int argc, char *argv[])
 		case 'q':
 			cf_quiet = 1;
 			break;
+		case 'u':
+			arg_username = optarg;
+			break;
 		case 'h':
+			usage(0, argv[0]);
 		default:
-			usage(1);
+			usage(1, argv[0]);
 		}
 	}
 	if (optind + 1 != argc)
-		usage(1);
+		usage(1, argv[0]);
 	cf_config_file = argv[optind];
+
+	init_objects();
 	load_config(false);
+	init_caches();
+
+	/* prefer cmdline over config for username */
+	if (arg_username)
+		cf_username = arg_username;
+
+	/* switch user is needed */
+	if (*cf_username)
+		change_user(cf_username);
+
+	/* disallow running as root */
+	if (getuid() == 0)
+		fatal("PgBouncer should not run as root");
 
 	/* need to do that after loading config */
 	check_limits();
 
-	/* init random */
-	srandom(time(NULL) ^ getpid());
+	admin_setup();
+
+	if (cf_reboot) {
+		if (check_old_process_unix()) {
+			takeover_part1();
+			did_takeover = true;
+		} else {
+			log_info("old process not found, try to continue normally");
+			cf_reboot = 0;
+			check_pidfile();
+		}
+	} else {
+		if (check_old_process_unix())
+			fatal("unix socket is in use, cannot continue");
+		check_pidfile();
+	}
+
+	if (cf_daemon)
+		go_daemon();
 
 	/* initialize subsystems, order important */
-	daemon_setup();
+	srandom(time(NULL) ^ getpid());
 	event_init();
 	signal_setup();
 	janitor_setup();
 	stats_setup();
-	admin_setup();
 
-	if (cf_reboot) {
-		takeover_init();
-		while (cf_reboot)
-			main_loop_once();
-		write_pidfile();
-	} else
+	if (did_takeover)
+		takeover_finish();
+	else
 		pooler_setup();
 
+	write_pidfile();
+
 	/* main loop */
 	while (1)
 		main_loop_once();
