@@ -27,15 +27,22 @@ STATLIST(user_list);
 STATLIST(database_list);
 STATLIST(pool_list);
 
+Tree user_tree;
+
 /*
  * client and server objects will be pre-allocated
  * they are always in either active or free lists
  * in addition to others.
  */
-STATLIST(free_client_list);
-STATLIST(free_server_list);
 STATLIST(login_client_list);
 
+ObjectCache *server_cache;
+ObjectCache *client_cache;
+ObjectCache *db_cache;
+ObjectCache *pool_cache;
+ObjectCache *user_cache;
+ObjectCache *iobuf_cache;
+
 /*
  * libevent may still report events when event_del()
  * is called from somewhere else.  So hide just freed
@@ -44,108 +51,70 @@ STATLIST(login_client_list);
 static STATLIST(justfree_client_list);
 static STATLIST(justfree_server_list);
 
-/* how many client sockets are allocated */
-static int absolute_client_count = 0;
-/* how many server sockets are allocated */
-static int absolute_server_count = 0;
-
-/* list of users ordered by name */
-static PgUser **user_lookup = NULL;
-
-/* drop lookup list because it will be out of sync */
-static void reset_auth_cache(void)
-{
-	if (user_lookup != NULL) {
-		free(user_lookup);
-		user_lookup = NULL;
-	}
-}
-
 /* fast way to get number of active clients */
 int get_active_client_count(void)
 {
-	return absolute_client_count - statlist_count(&free_client_list);
+	return objcache_active_count(client_cache);
 }
 
 /* fast way to get number of active servers */
 int get_active_server_count(void)
 {
-	return absolute_server_count - statlist_count(&free_server_list);
+	return objcache_active_count(server_cache);
 }
 
-/* this should be called on free socket that is put into use */
-static void clean_socket(PgSocket *sk)
+static void construct_client(void *obj)
 {
-	sk->link = NULL;
-	sk->pool = NULL;
-
-	sk->wait_for_welcome = 0;
-	sk->ready = 0;
-	sk->admin_user = 0;
-	sk->own_user = 0;
-	sk->suspended = 0;
-	sk->wait_for_response = 0;
-
-	sk->connect_time = 0;
-	sk->request_time = 0;
-	sk->query_start = 0;
-
-	sk->auth_user = NULL;
-
-	varcache_clean(&sk->vars);
-	sk->setting_vars = 0;
-}
-
-/* allocate & fill client socket */
-static PgSocket *new_client(void)
-{
-	PgSocket *client;
-
-	/* get free PgSocket */
-	client = first_socket(&free_client_list);
-	if (client) {
-		clean_socket(client);
-		return client;
-	}
-
-	client = zmalloc(sizeof(*client) + cf_sbuf_len);
-	if (!client)
-		return NULL;
+	PgSocket *client = obj;
 
+	memset(client, 0, sizeof(PgSocket));
 	list_init(&client->head);
-	sbuf_init(&client->sbuf, client_proto, client);
-	statlist_prepend(&client->head, &free_client_list);
+	sbuf_init(&client->sbuf, client_proto);
 	client->state = CL_FREE;
-
-	absolute_client_count++;
-
-	return client;
 }
 
-/* allocate & fill server socket */
-static PgSocket *new_server(void)
+static void construct_server(void *obj)
 {
-	PgSocket *server;
-
-	/* get free PgSocket */
-	server = first_socket(&free_server_list);
-	if (server) {
-		clean_socket(server);
-		return server;
-	}
-
-	server = zmalloc(sizeof(*server) + cf_sbuf_len);
-	if (!server)
-		return NULL;
+	PgSocket *server = obj;
 
+	memset(server, 0, sizeof(PgSocket));
 	list_init(&server->head);
-	sbuf_init(&server->sbuf, server_proto, server);
-	statlist_prepend(&server->head, &free_server_list);
+	sbuf_init(&server->sbuf, server_proto);
 	server->state = SV_FREE;
+}
 
-	absolute_server_count++;
+/* compare string with PgUser->name, for usage with btree */
+static int user_node_cmp(long userptr, Node *node)
+{
+	const char *name = (const char *)userptr;
+	PgUser *user = container_of(node, PgUser, tree_node);
+	return strcmp(name, user->name);
+}
 
-	return server;
+/* initialization before config loading */
+void init_objects(void)
+{
+	tree_init(&user_tree, user_node_cmp, NULL);
+	user_cache = objcache_create("user_cache", sizeof(PgUser), 0, NULL);
+	db_cache = objcache_create("db_cache", sizeof(PgDatabase), 0, NULL);
+	pool_cache = objcache_create("pool_cache", sizeof(PgPool), 0, NULL);
+
+	if (!user_cache || !db_cache || !pool_cache)
+		fatal("cannot create initial caches");
+}
+
+static void do_iobuf_reset(void *arg)
+{
+	IOBuf *io = arg;
+	iobuf_reset(io);
+}
+
+/* initialization after config loading */
+void init_caches(void)
+{
+	server_cache = objcache_create("server_cache", sizeof(PgSocket), 0, construct_server);
+	client_cache = objcache_create("client_cache", sizeof(PgSocket), 0, construct_client);
+	iobuf_cache = objcache_create("iobuf_cache", IOBUF_SIZE, 0, do_iobuf_reset);
 }
 
 /* state change means moving between lists */
@@ -156,7 +125,6 @@ void change_client_state(PgSocket *client, SocketState newstate)
 	/* remove from old location */
 	switch (client->state) {
 	case CL_FREE:
-		statlist_remove(&client->head, &free_client_list);
 		break;
 	case CL_JUSTFREE:
 		statlist_remove(&client->head, &justfree_client_list);
@@ -182,7 +150,7 @@ void change_client_state(PgSocket *client, SocketState newstate)
 	/* put to new location */
 	switch (client->state) {
 	case CL_FREE:
-		statlist_prepend(&client->head, &free_client_list);
+		obj_free(client_cache, client);
 		break;
 	case CL_JUSTFREE:
 		statlist_append(&client->head, &justfree_client_list);
@@ -212,7 +180,6 @@ void change_server_state(PgSocket *server, SocketState newstate)
 	/* remove from old location */
 	switch (server->state) {
 	case SV_FREE:
-		statlist_remove(&server->head, &free_server_list);
 		break;
 	case SV_JUSTFREE:
 		statlist_remove(&server->head, &justfree_server_list);
@@ -241,7 +208,7 @@ void change_server_state(PgSocket *server, SocketState newstate)
 	/* put to new location */
 	switch (server->state) {
 	case SV_FREE:
-		statlist_prepend(&server->head, &free_server_list);
+		obj_free(server_cache, server);
 		break;
 	case SV_JUSTFREE:
 		statlist_append(&server->head, &justfree_server_list);
@@ -325,12 +292,12 @@ PgDatabase *add_database(const char *name)
 
 	/* create new object if needed */
 	if (db == NULL) {
-		db = zmalloc(sizeof(*db));
+		db = obj_alloc(db_cache);
 		if (!db)
 			return NULL;
 
 		list_init(&db->head);
-		strlcpy(db->name, name, sizeof(db->name));
+		safe_strcpy(db->name, name, sizeof(db->name));
 		put_in_order(&db->head, &database_list, cmp_database);
 	}
 
@@ -342,19 +309,19 @@ PgUser *add_user(const char *name, const char *passwd)
 {
 	PgUser *user = find_user(name);
 
-	reset_auth_cache();
-
 	if (user == NULL) {
-		user = zmalloc(sizeof(*user));
+		user = obj_alloc(user_cache);
 		if (!user)
 			return NULL;
 
 		list_init(&user->head);
 		list_init(&user->pool_list);
-		strlcpy(user->name, name, sizeof(user->name));
+		safe_strcpy(user->name, name, sizeof(user->name));
 		put_in_order(&user->head, &user_list, cmp_user);
+
+		tree_insert(&user_tree, (long)user->name, &user->tree_node);
 	}
-	strlcpy(user->passwd, passwd, sizeof(user->passwd));
+	safe_strcpy(user->passwd, passwd, sizeof(user->passwd));
 	return user;
 }
 
@@ -363,14 +330,14 @@ PgUser *force_user(PgDatabase *db, const char *name, const char *passwd)
 {
 	PgUser *user = db->forced_user;
 	if (!user) {
-		user = zmalloc(sizeof(*user));
+		user = obj_alloc(user_cache);
 		if (!user)
 			return NULL;
 		list_init(&user->head);
 		list_init(&user->pool_list);
 	}
-	strlcpy(user->name, name, sizeof(user->name));
-	strlcpy(user->passwd, passwd, sizeof(user->passwd));
+	safe_strcpy(user->name, name, sizeof(user->name));
+	safe_strcpy(user->passwd, passwd, sizeof(user->passwd));
 	db->forced_user = user;
 	return user;
 }
@@ -388,56 +355,15 @@ PgDatabase *find_database(const char *name)
 	return NULL;
 }
 
-/* compare string with PgUser->name, for usage with bsearch() */
-static int user_name_cmp(const void *namestr, const void *userptr)
-{
-	const PgUser * const *user_p = userptr;
-	const PgUser *user = *user_p;
-	return strcmp(namestr, user->name);
-}
-
 /* find existing user */
 PgUser *find_user(const char *name)
 {
-	List *item;
-	PgUser *user;
+	PgUser *user = NULL;
+	Node *node;
 
-	/* if lookup table is available, use faster method */
-	if (user_lookup) {
-		PgUser **res;
-		res = bsearch(name, user_lookup,
-			      statlist_count(&user_list),
-			      sizeof(PgUser *),
-			      user_name_cmp);
-		return res ? *res : NULL;
-	}
-
-	/* slow lookup */
-	statlist_for_each(item, &user_list) {
-		user = container_of(item, PgUser, head);
-		if (strcmp(user->name, name) == 0)
-			return user;
-	}
-	return NULL;
-}
-
-/* create lookup list */
-void create_auth_cache(void)
-{
-	int i = 0;
-	List *item;
-	PgUser *user;
-
-	reset_auth_cache();
-
-	user_lookup = malloc(sizeof(PgUser *) * statlist_count(&user_list));
-	if (!user_lookup)
-		return;
-
-	statlist_for_each(item, &user_list) {
-		user = container_of(item, PgUser, head);
-		user_lookup[i++] = user;
-	}
+	node = tree_search(&user_tree, (long)name);
+	user = node ? container_of(node, PgUser, tree_node) : NULL;
+	return user;
 }
 
 /* create new pool object */
@@ -445,7 +371,7 @@ static PgPool *new_pool(PgDatabase *db, PgUser *user)
 {
 	PgPool *pool;
 
-	pool = zmalloc(sizeof(*pool));
+	pool = obj_alloc(pool_cache);
 	if (!pool)
 		return NULL;
 
@@ -491,13 +417,14 @@ PgPool *get_pool(PgDatabase *db, PgUser *user)
 }
 
 /* deactivate socket and put into wait queue */
-void pause_client(PgSocket *client)
+static void pause_client(PgSocket *client)
 {
 	Assert(client->state == CL_ACTIVE);
 
 	slog_debug(client, "pause_client");
 	change_client_state(client, CL_WAITING);
-	sbuf_pause(&client->sbuf);
+	if (!sbuf_pause(&client->sbuf))
+		disconnect_client(client, true, "pause failed");
 }
 
 /* wake client from wait */
@@ -551,15 +478,15 @@ bool find_server(PgSocket *client)
 		server->link = client;
 		change_server_state(server, SV_ACTIVE);
 		if (varchange) {
-			sbuf_pause(&client->sbuf);
-			res = false; /* don't process client data yet */
 			server->setting_vars = 1;
 			server->ready = 0;
+			res = false; /* don't process client data yet */
+			if (!sbuf_pause(&client->sbuf))
+				disconnect_client(client, true, "pause failed");
 		} else
 			res = true;
 	} else {
 		pause_client(client);
-		Assert(client->state == CL_WAITING);
 		res = false;
 	}
 	return res;
@@ -599,6 +526,26 @@ static bool reset_on_release(PgSocket *server)
 	return res;
 }
 
+static bool life_over(PgSocket *server)
+{
+	PgPool *pool = server->pool;
+	usec_t lifetime_kill_gap = 0;
+	usec_t now = get_cached_time();
+	usec_t age = now - server->connect_time;
+	usec_t last_kill = now - pool->last_lifetime_disconnect;
+
+	if (age < cf_server_lifetime)
+		return false;
+
+	if (pool->db->pool_size > 0)
+		lifetime_kill_gap = cf_server_lifetime / pool->db->pool_size;
+
+	if (last_kill >= lifetime_kill_gap)
+		return true;
+
+	return false;
+}
+
 /* connecting/active -> idle, unlink if needed */
 bool release_server(PgSocket *server)
 {
@@ -633,6 +580,18 @@ bool release_server(PgSocket *server)
 		fatal("bad server state in release_server");
 	}
 
+	/* enforce lifetime immidiately on release */
+	if (server->state != SV_LOGIN && life_over(server)) {
+		disconnect_server(server, true, "server_lifetime");
+		return false;
+	}
+
+	/* enforce close request */
+	if (server->close_needed) {
+		disconnect_server(server, true, "close_needed");
+		return false;
+	}
+
 	Assert(server->link == NULL);
 	slog_noise(server, "release_server: new state=%d", newstate);
 	change_server_state(server, newstate);
@@ -651,7 +610,7 @@ void disconnect_server(PgSocket *server, bool notify, const char *reason)
 {
 	PgPool *pool = server->pool;
 	PgSocket *client = server->link;
-	static const uint8 pkt_term[] = {'X', 0,0,0,4};
+	static const uint8_t pkt_term[] = {'X', 0,0,0,4};
 	int send_term = 1;
 	usec_t now = get_cached_time();
 
@@ -689,11 +648,15 @@ void disconnect_server(PgSocket *server, bool notify, const char *reason)
 	Assert(server->link == NULL);
 
 	/* notify server and close connection */
-	if (send_term && notify)
-		sbuf_answer(&server->sbuf, pkt_term, sizeof(pkt_term));
-	sbuf_close(&server->sbuf);
+	if (send_term && notify) {
+		if (!sbuf_answer(&server->sbuf, pkt_term, sizeof(pkt_term)))
+			/* ignore result */
+			notify = false;
+	}
 
 	change_server_state(server, SV_JUSTFREE);
+	if (!sbuf_close(&server->sbuf))
+		log_noise("sbuf_close failed, retry later");
 }
 
 /* drop client connection */
@@ -711,6 +674,7 @@ void disconnect_client(PgSocket *client, bool notify, const char *reason)
 			PgSocket *server = client->link;
 			/* ->ready may be set before all is sent */
 			if (server->ready && sbuf_is_empty(&server->sbuf)) {
+				/* retval does not matter here */
 				release_server(server);
 			} else {
 				server->link = NULL;
@@ -727,7 +691,7 @@ void disconnect_client(PgSocket *client, bool notify, const char *reason)
 	}
 
 	/* send reason to client */
-	if (notify && reason) {
+	if (notify && reason && client->state != CL_CANCEL) {
 		/*
 		 * don't send Ready pkt here, or client won't notice
 		 * closed connection
@@ -735,9 +699,9 @@ void disconnect_client(PgSocket *client, bool notify, const char *reason)
 		send_pooler_error(client, false, reason);
 	}
 
-	sbuf_close(&client->sbuf);
-
 	change_client_state(client, CL_JUSTFREE);
+	if (!sbuf_close(&client->sbuf))
+		log_noise("sbuf_close failed, retry later");
 }
 
 /* the pool needs new connection, if possible */
@@ -746,6 +710,7 @@ void launch_new_connection(PgPool *pool)
 	PgSocket *server;
 	int total;
 	const char *unix_dir = cf_unix_socket_dir;
+	bool res;
 
 	/* allow only small number of connection attempts at a time */
 	if (!statlist_empty(&pool->new_server_list)) {
@@ -771,7 +736,7 @@ void launch_new_connection(PgPool *pool)
 	}
 
 	/* get free conn object */
-	server = new_server();
+	server = obj_alloc(server_cache);
 	if (!server) {
 		log_debug("launch_new_connection: no memory");
 		return;
@@ -793,8 +758,10 @@ void launch_new_connection(PgPool *pool)
 		unix_dir = server->pool->db->unix_socket_dir;
 
 	/* start connecting */
-	sbuf_connect(&server->sbuf, &server->remote_addr, unix_dir,
-		     cf_server_connect_timeout / USEC);
+	res = sbuf_connect(&server->sbuf, &server->remote_addr, unix_dir,
+			   cf_server_connect_timeout / USEC);
+	if (!res)
+		log_noise("failed to launch new connection");
 }
 
 /* new client connection attempt */
@@ -802,12 +769,16 @@ PgSocket * accept_client(int sock,
 			 const struct sockaddr_in *addr,
 			 bool is_unix)
 {
+	bool res;
 	PgSocket *client;
 
 	/* get free PgSocket */
-	client = new_client();
-	if (!client)
+	client = obj_alloc(client_cache);
+	if (!client) {
+		log_warning("cannot allocate client struct");
+		safe_close(sock);
 		return NULL;
+	}
 
 	client->connect_time = client->request_time = get_cached_time();
 	client->query_start = 0;
@@ -817,10 +788,12 @@ PgSocket * accept_client(int sock,
 
 	change_client_state(client, CL_LOGIN);
 
-	if (cf_log_connections)
-		slog_debug(client, "got connection attempt");
-	sbuf_accept(&client->sbuf, sock, is_unix);
-
+	res = sbuf_accept(&client->sbuf, sock, is_unix);
+	if (!res) {
+		if (cf_log_connections)
+			slog_debug(client, "failed connection attempt");
+		return NULL;
+	}
 
 	return client;
 }
@@ -850,10 +823,6 @@ bool finish_client_login(PgSocket *client)
 
 	slog_debug(client, "logged in");
 
-	/* in suspend, don't let send query */
-	if (cf_pause_mode == P_SUSPEND)
-		suspend_socket(client);
-
 	return true;
 }
 
@@ -888,14 +857,16 @@ found:
 	/* not linked client, just drop it then */
 	if (!main_client->link) {
 		bool res;
-		disconnect_client(req, false, "cancel request for idle client");
 
 		/* let administrative cancel be handled elsewhere */
-		if (main_client->pool->admin) {
+		if (main_client->pool->db->admin) {
+			disconnect_client(req, false, "cancel request for console client");
 			admin_handle_cancel(main_client);
 			return;
 		}
 
+		disconnect_client(req, false, "cancel request for idle client");
+
 		/* notify readiness */
 		SEND_ReadyForQuery(res, main_client);
 		if (!res)
@@ -904,8 +875,9 @@ found:
 		return;
 	}
 
-	/* drop the connection silently */
-	sbuf_close(&req->sbuf);
+	/* drop the connection, if fails, retry later in justfree list */
+	if (!sbuf_close(&req->sbuf))
+		log_noise("sbuf_close failed, retry later");
 
 	/* remember server key */
 	server = main_client->link;
@@ -934,7 +906,7 @@ void forward_cancel_request(PgSocket *server)
 
 bool use_client_socket(int fd, PgAddr *addr,
 		       const char *dbname, const char *username,
-		       uint64 ckey, int oldfd, int linkfd,
+		       uint64_t ckey, int oldfd, int linkfd,
 		       const char *client_enc, const char *std_string,
 		       const char *datestyle, const char *timezone)
 {
@@ -942,6 +914,8 @@ bool use_client_socket(int fd, PgAddr *addr,
 	PktBuf tmp;
 
 	client = accept_client(fd, NULL, addr->is_unix);
+	if (client == NULL)
+		return false;
 	client->suspended = 1;
 
 	if (!set_pool(client, dbname, username))
@@ -967,7 +941,7 @@ bool use_client_socket(int fd, PgAddr *addr,
 
 bool use_server_socket(int fd, PgAddr *addr,
 		       const char *dbname, const char *username,
-		       uint64 ckey, int oldfd, int linkfd,
+		       uint64_t ckey, int oldfd, int linkfd,
 		       const char *client_enc, const char *std_string,
 		       const char *datestyle, const char *timezone)
 {
@@ -976,6 +950,7 @@ bool use_server_socket(int fd, PgAddr *addr,
 	PgPool *pool;
 	PgSocket *server;
 	PktBuf tmp;
+	bool res;
 
 	if (db->forced_user)
 		user = db->forced_user;
@@ -986,11 +961,14 @@ bool use_server_socket(int fd, PgAddr *addr,
 	if (!pool)
 		return false;
 
-	server = new_server();
+	server = obj_alloc(server_cache);
 	if (!server)
 		return false;
 
-	sbuf_accept(&server->sbuf, fd, addr->is_unix);
+	res = sbuf_accept(&server->sbuf, fd, addr->is_unix);
+	if (!res)
+		return false;
+
 	server->suspended = 1;
 	server->pool = pool;
 	server->auth_user = user;
@@ -1066,19 +1044,28 @@ void reuse_just_freed_objects(void)
 {
 	List *tmp, *item;
 	PgSocket *sk;
+	bool close_works = true;
 
 	/*
-	 * Obviously, if state would be set to *_FREE,
-	 * they could be moved in one go.
+	 * event_del() may fail because of ENOMEM for event handlers
+	 * that need only changes sent to kernel on each loop.
+	 *
+	 * Keep open sbufs in justfree lists until successful.
 	 */
+
 	statlist_for_each_safe(item, &justfree_client_list, tmp) {
 		sk = container_of(item, PgSocket, head);
-		change_client_state(sk, CL_FREE);
+		if (sbuf_is_closed(&sk->sbuf))
+			change_client_state(sk, CL_FREE);
+		else if (close_works)
+			close_works = sbuf_close(&sk->sbuf);
 	}
 	statlist_for_each_safe(item, &justfree_server_list, tmp) {
 		sk = container_of(item, PgSocket, head);
-		change_server_state(sk, SV_FREE);
+		if (sbuf_is_closed(&sk->sbuf))
+			change_server_state(sk, SV_FREE);
+		else if (close_works)
+			close_works = sbuf_close(&sk->sbuf);
 	}
 }
 
-
