@@ -55,6 +55,26 @@ static bool load_parameter(PgSocket *server, PktHdr *pkt, bool startup)
 	return true;
 }
 
+/* we cannot log in at all, notify clients */
+static void kill_pool_logins(PgPool *pool, PktHdr *errpkt)
+{
+	List *item, *tmp;
+	PgSocket *client;
+	const char *level, *msg;
+
+	parse_server_error(errpkt, &level, &msg);
+
+	log_warning("server login failed: %s %s", level, msg);
+
+	statlist_for_each_safe(item, &pool->waiting_client_list, tmp) {
+		client = container_of(item, PgSocket, head);
+		if (!client->wait_for_welcome)
+			continue;
+
+		disconnect_client(client, true, msg);
+	}
+}
+
 /* process packets on server auth phase */
 static bool handle_server_startup(PgSocket *server, PktHdr *pkt)
 {
@@ -66,6 +86,20 @@ static bool handle_server_startup(PgSocket *server, PktHdr *pkt)
 		return false;
 	}
 
+	/* ignore most that happens during connect_query */
+	if (server->exec_on_connect) {
+		switch (pkt->type) {
+		case 'Z':
+		case 'S':	/* handle them below */
+			break;
+
+		case 'E':	/* log & ignore errors */
+			log_server_error("S: error while executing exec_on_query", pkt);
+		default:	/* ignore rest */
+			sbuf_prepare_skip(sbuf, pkt->len);
+			return true;
+		}
+	}
 
 	switch (pkt->type) {
 	default:
@@ -74,7 +108,11 @@ static bool handle_server_startup(PgSocket *server, PktHdr *pkt)
 		break;
 
 	case 'E':		/* ErrorResponse */
-		log_server_error("S: login failed", pkt);
+		if (!server->pool->welcome_msg_ready)
+			kill_pool_logins(server->pool, pkt);
+		else
+			log_server_error("S: login failed", pkt);
+
 		disconnect_server(server, true, "login failed");
 		break;
 
@@ -91,6 +129,18 @@ static bool handle_server_startup(PgSocket *server, PktHdr *pkt)
 		break;
 
 	case 'Z':		/* ReadyForQuery */
+		if (server->exec_on_connect) {
+			server->exec_on_connect = 0;
+			/* deliberately ignore transaction status */
+		} else if (server->pool->db->connect_query) {
+			server->exec_on_connect = 1;
+			slog_debug(server, "server conect ok, send exec_on_connect");
+			SEND_generic(res, server, 'Q', "s", server->pool->db->connect_query);
+			if (!res)
+				disconnect_server(server, false, "exec_on_connect query failed");
+			break;
+		}
+
 		/* login ok */
 		slog_debug(server, "server login ok, start accepting queries");
 		server->ready = 1;
@@ -98,10 +148,11 @@ static bool handle_server_startup(PgSocket *server, PktHdr *pkt)
 		/* got all params */
 		finish_welcome_msg(server);
 
+		/* need to notify sbuf if server was closed */
 		res = release_server(server);
 
 		/* let the takeover process handle it */
-		if (res && server->pool->admin)
+		if (res && server->pool->db->admin)
 			res = takeover_login(server);
 		break;
 
@@ -134,7 +185,7 @@ static bool handle_server_work(PgSocket *server, PktHdr *pkt)
 	SBuf *sbuf = &server->sbuf;
 	PgSocket *client = server->link;
 
-	Assert(!server->pool->admin);
+	Assert(!server->pool->db->admin);
 
 	switch (pkt->type) {
 	default:
@@ -191,6 +242,7 @@ static bool handle_server_work(PgSocket *server, PktHdr *pkt)
 			disconnect_server(server, true, "invalid server parameter");
 			return false;
 		}
+
 	case 'N':		/* NoticeResponse */
 		break;
 
@@ -269,20 +321,19 @@ static bool handle_connect(PgSocket *server)
 }
 
 /* callback from SBuf */
-bool server_proto(SBuf *sbuf, SBufEvent evtype, MBuf *data, void *arg)
+bool server_proto(SBuf *sbuf, SBufEvent evtype, MBuf *data)
 {
 	bool res = false;
-	PgSocket *server = arg;
+	PgSocket *server = container_of(sbuf, PgSocket, sbuf);
+	PgPool *pool = server->pool;
 	PktHdr pkt;
 
 	Assert(is_server_socket(server));
 	Assert(server->state != SV_FREE);
 
-	if (server->state == SV_JUSTFREE) {
-		/* SBuf should catch the case */
-		slog_warning(server, "state=SV_JUSTFREE, should not happen");
+	/* may happen if close failed */
+	if (server->state == SV_JUSTFREE)
 		return false;
-	}
 
 	switch (evtype) {
 	case SBUF_EV_RECV_FAILED:
@@ -294,13 +345,13 @@ bool server_proto(SBuf *sbuf, SBufEvent evtype, MBuf *data, void *arg)
 	case SBUF_EV_READ:
 		if (mbuf_avail(data) < NEW_HEADER_LEN) {
 			slog_noise(server, "S: got partial header, trying to wait a bit");
-			return false;
+			break;
 		}
 
 		/* parse pkt header */
 		if (!get_header(data, &pkt)) {
 			disconnect_server(server, true, "bad pkt header");
-			return false;
+			break;
 		}
 		slog_noise(server, "S: pkt '%c', len=%d", pkt_desc(&pkt), pkt.len);
 
@@ -330,7 +381,7 @@ bool server_proto(SBuf *sbuf, SBufEvent evtype, MBuf *data, void *arg)
 		res = handle_connect(server);
 		break;
 	case SBUF_EV_FLUSH:
-		res = true; /* unused actually */
+		res = true;
 		if (!server->ready)
 			break;
 
@@ -357,7 +408,12 @@ bool server_proto(SBuf *sbuf, SBufEvent evtype, MBuf *data, void *arg)
 			}
 		}
 		break;
+	case SBUF_EV_PKT_CALLBACK:
+		slog_warning(server, "SBUF_EV_PKT_CALLBACK with state=%d", server->state);
+		break;
 	}
+	if (!res && pool->db->admin)
+		takeover_login_failed();
 	return res;
 }
 
