@@ -39,7 +39,7 @@ static void stat_add(PgStats *total, PgStats *stat)
 
 static void calc_average(PgStats *avg, PgStats *cur, PgStats *old)
 {
-	uint64 qcount;
+	uint64_t qcount;
 	usec_t dur = get_cached_time() - old_stamp;
 
 	reset_stats(avg);
@@ -122,6 +122,49 @@ bool admin_database_stats(PgSocket *client, StatList *pool_list)
 	return true;
 }
 
+bool show_stat_totals(PgSocket *client, StatList *pool_list)
+{
+	PgPool *pool;
+	List *item;
+	PgStats st_total, old_total, avg;
+	PktBuf *buf;
+
+	reset_stats(&st_total);
+	reset_stats(&old_total);
+
+	buf = pktbuf_dynamic(512);
+	if (!buf) {
+		admin_error(client, "no mem");
+		return true;
+	}
+
+
+	statlist_for_each(item, pool_list) {
+		pool = container_of(item, PgPool, head);
+		stat_add(&st_total, &pool->stats);
+		stat_add(&old_total, &pool->older_stats);
+	}
+
+	calc_average(&avg, &st_total, &old_total);
+
+	pktbuf_write_RowDescription(buf, "sq", "name", "value");
+
+#define WTOTAL(name) pktbuf_write_DataRow(buf, "sq", "total_" #name, st_total.name)
+#define WAVG(name) pktbuf_write_DataRow(buf, "sq", "avg_" #name, avg.name)
+
+	WTOTAL(request_count);
+	WTOTAL(client_bytes);
+	WTOTAL(server_bytes);
+	WTOTAL(query_time);
+	WAVG(request_count);
+	WAVG(client_bytes);
+	WAVG(server_bytes);
+	WAVG(query_time);
+
+	admin_flush(client, buf, "SHOW");
+	return true;
+}
+
 static void refresh_stats(int s, short flags, void *arg)
 {
 	List *item;
@@ -143,14 +186,16 @@ static void refresh_stats(int s, short flags, void *arg)
 		stat_add(&cur_total, &pool->stats);
 		stat_add(&old_total, &pool->older_stats);
 	}
-	evtimer_add(&ev_stats, &period);
-
 	calc_average(&avg, &cur_total, &old_total);
 	/* send totals to logfile */
-	log_info("Stats: %llu req/s, in %llu b/s, "
-		 "out %llu b/s, query %llu us",
+	log_info("Stats: %" PRIu64 " req/s,"
+		 " in %" PRIu64 " b/s,"
+		 " out %" PRIu64 " b/s,"
+		 "query %" PRIu64 " us",
 		 avg.request_count, avg.client_bytes,
 		 avg.server_bytes, avg.query_time);
+
+	safe_evtimer_add(&ev_stats, &period);
 }
 
 void stats_setup(void)
@@ -160,8 +205,8 @@ void stats_setup(void)
 	new_stamp = get_cached_time();
 	old_stamp = new_stamp - USEC;
 
-	/* launch maintenance */
+	/* launch stats */
 	evtimer_set(&ev_stats, refresh_stats, NULL);
-	evtimer_add(&ev_stats, &period);
+	safe_evtimer_add(&ev_stats, &period);
 }
 
