@@ -31,24 +31,60 @@
  * Takeover done, old process shut down,
  * kick this one running.
  */
-static void takeover_finish(PgSocket *bouncer)
+
+static PgSocket *old_bouncer = NULL;
+
+void takeover_finish(void)
 {
-	disconnect_server(bouncer, false, "disko over");
-	cf_reboot = 0;
+	uint8_t buf[512];
+	int fd = sbuf_socket(&old_bouncer->sbuf);
+	bool res;
+	int got;
+
+	log_info("sending SHUTDOWN;");
+	socket_set_nonblocking(fd, 0);
+	SEND_generic(res, old_bouncer, 'Q', "s", "SHUTDOWN;");
+	if (!res)
+		fatal("failed to send SHUTDOWN;");
+
+	while (1) {
+		got = safe_recv(fd, buf, sizeof(buf), 0);
+		if (got == 0)
+			break;
+		if (got < 0)
+			fatal_perror("sky is falling - error while waiting result from SHUTDOWN");
+	}
+
+	disconnect_server(old_bouncer, false, "disko over");
+	old_bouncer = NULL;
+
+	log_info("old process killed, resuming work");
 	resume_all();
-	log_info("disko over, resuming work");
+}
+
+static void takeover_finish_part1(PgSocket *bouncer)
+{
+	Assert(old_bouncer == NULL);
+
+	/* unregister bouncer from libevent */
+	if (!sbuf_pause(&bouncer->sbuf))
+		fatal_perror("sbuf_pause failed");
+	old_bouncer = bouncer;
+	cf_reboot = 0;
+	log_info("disko over, going background");
 }
 
 /* parse msg for fd and info */
-static bool takeover_load_fd(MBuf *pkt, const struct cmsghdr *cmsg)
+static void takeover_load_fd(MBuf *pkt, const struct cmsghdr *cmsg)
 {
 	int fd;
-	char *task, *s_addr, *user, *db;
+	char *task, *saddr, *user, *db;
 	char *client_enc, *std_string, *datestyle, *timezone;
 	int oldfd, port, linkfd;
 	int got;
-	uint64 ckey;
+	uint64_t ckey;
 	PgAddr addr;
+	bool res = false;
 
 	memset(&addr, 0, sizeof(addr));
 
@@ -64,9 +100,9 @@ static bool takeover_load_fd(MBuf *pkt, const struct cmsghdr *cmsg)
 
 	/* parse row contents */
 	got = scan_text_result(pkt, "issssiqissss", &oldfd, &task, &user, &db,
-			       &s_addr, &port, &ckey, &linkfd,
+			       &saddr, &port, &ckey, &linkfd,
 			       &client_enc, &std_string, &datestyle, &timezone);
-	if (task == NULL || s_addr == NULL)
+	if (task == NULL || saddr == NULL)
 		fatal("NULL data from old process");
 
 	log_debug("FD row: fd=%d(%d) linkfd=%d task=%s user=%s db=%s enc=%s",
@@ -75,27 +111,28 @@ static bool takeover_load_fd(MBuf *pkt, const struct cmsghdr *cmsg)
 		  client_enc ? client_enc : "NULL");
 
 	/* fill address */
-	addr.is_unix = strcmp(s_addr, "unix") == 0 ? true : false;
+	addr.is_unix = strcmp(saddr, "unix") == 0 ? true : false;
 	if (addr.is_unix) {
 		addr.port = cf_listen_port;
 	} else {
-		addr.ip_addr.s_addr = inet_addr(s_addr);
+		addr.ip_addr.s_addr = inet_addr(saddr);
 		addr.port = port;
 	}
 
 	/* decide what to do with it */
 	if (strcmp(task, "client") == 0)
-		use_client_socket(fd, &addr, db, user, ckey, oldfd, linkfd,
+		res = use_client_socket(fd, &addr, db, user, ckey, oldfd, linkfd,
 				  client_enc, std_string, datestyle, timezone);
 	else if (strcmp(task, "server") == 0)
-		use_server_socket(fd, &addr, db, user, ckey, oldfd, linkfd,
+		res = use_server_socket(fd, &addr, db, user, ckey, oldfd, linkfd,
 				  client_enc, std_string, datestyle, timezone);
 	else if (strcmp(task, "pooler") == 0)
-		use_pooler_socket(fd, addr.is_unix);
+		res = use_pooler_socket(fd, addr.is_unix);
 	else
 		fatal("unknown task: %s", task);
 
-	return true;
+	if (!res)
+		fatal("socket takeover failed - no mem?");
 }
 
 static void takeover_create_link(PgPool *pool, PgSocket *client)
@@ -137,7 +174,7 @@ static void takeover_postprocess_fds(void)
 
 	statlist_for_each(item, &pool_list) {
 		pool = container_of(item, PgPool, head);
-		if (pool->admin)
+		if (pool->db->admin)
 			continue;
 		statlist_for_each(item2, &pool->active_client_list) {
 			client = container_of(item2, PgSocket, head);
@@ -163,14 +200,11 @@ static void next_command(PgSocket *bouncer, MBuf *pkt)
 		log_info("SUSPEND finished, sending SHOW FDS");
 		SEND_generic(res, bouncer, 'Q', "s", "SHOW FDS;");
 	} else if (strncmp(cmd, "SHOW", 4) == 0) {
-
-		log_info("SHOW FDS finished, sending SHUTDOWN");
-
 		/* all fds loaded, review them */
 		takeover_postprocess_fds();
+		log_info("SHOW FDS finished");
 
-		/* all OK, kill old one */
-		SEND_generic(res, bouncer, 'Q', "s", "SHUTDOWN;");
+		takeover_finish_part1(bouncer);
 	} else
 		fatal("got bad CMD from old bouncer: %s", cmd);
 
@@ -227,13 +261,13 @@ static void takeover_parse_data(PgSocket *bouncer,
 /*
  * listen for data from old bouncer.
  *
- * use always sendmsg, to keep code simpler
+ * use always recvmsg, to keep code simpler
  */
 static void takeover_recv_cb(int sock, short flags, void *arg)
 {
-	PgSocket *bouncer = arg;
-	uint8 data_buf[2048];
-	uint8 cnt_buf[128];
+	PgSocket *bouncer = container_of(arg, PgSocket, sbuf);
+	uint8_t data_buf[2048];
+	uint8_t cnt_buf[128];
 	struct msghdr msg;
 	struct iovec io;
 	int res;
@@ -252,7 +286,7 @@ static void takeover_recv_cb(int sock, short flags, void *arg)
 		mbuf_init(&data, data_buf, res);
 		takeover_parse_data(bouncer, &msg, &data);
 	} else if (res == 0) {
-		takeover_finish(bouncer);
+		fatal("unexpected EOF");
 	} else {
 		if (errno == EAGAIN)
 			return;
@@ -272,10 +306,13 @@ bool takeover_login(PgSocket *bouncer)
 	SEND_generic(res, bouncer, 'Q', "s", "SUSPEND;");
 	if (res) {
 		/* use own callback */
-		sbuf_pause(&bouncer->sbuf);
-		sbuf_continue_with_callback(&bouncer->sbuf, takeover_recv_cb);
+		if (!sbuf_pause(&bouncer->sbuf))
+			fatal("sbuf_pause failed");
+		res = sbuf_continue_with_callback(&bouncer->sbuf, takeover_recv_cb);
+		if (!res)
+			fatal("takeover_login: sbuf_continue_with_callback failed");
 	} else {
-		disconnect_server(bouncer, false, "failed to send command");
+		fatal("takeover_login: failed to send command");
 	}
 	return res;
 }
@@ -293,3 +330,8 @@ void takeover_init(void)
 	launch_new_connection(pool);
 }
 
+void takeover_login_failed(void)
+{
+	fatal("login failed");
+}
+
