@@ -22,8 +22,33 @@
 
 #include "bouncer.h"
 
+#include <syslog.h>
+
 #include "md5.h"
 
+static int syslog_started = 0;
+static int log_fd = 0;
+
+struct FacName { const char *name; int code; };
+static struct FacName facility_names [] = {
+	{ "auth",	LOG_AUTH },
+#ifdef LOG_AUTHPRIV
+        { "authpriv",	LOG_AUTHPRIV },
+#endif
+	{ "daemon",	LOG_DAEMON },
+	{ "user",	LOG_USER },
+	{ "local0",	LOG_LOCAL0 },
+	{ "local1",	LOG_LOCAL1 },
+	{ "local2",	LOG_LOCAL2 },
+	{ "local3",	LOG_LOCAL3 },
+	{ "local4",	LOG_LOCAL4 },
+	{ "local5",	LOG_LOCAL5 },
+	{ "local6",	LOG_LOCAL6 },
+	{ "local7",	LOG_LOCAL7 },
+	{ NULL },
+};
+
+
 void *zmalloc(size_t len)
 {
 	void *p = malloc(len);
@@ -45,7 +70,48 @@ static void render_time(char *buf, int max)
 	strftime(buf, max, "%Y-%m-%d %H:%M:%S", &tm);
 }
 
-static int log_fd = 0;
+static void close_syslog(void)
+{
+	if (syslog_started) {
+		closelog();
+		syslog_started = 0;
+	}
+}
+
+static void init_syslog(void)
+{
+	struct FacName *fn;
+	int facility = LOG_DAEMON;
+
+	for (fn = facility_names; fn->name; fn++)
+		if (strcmp(cf_syslog_facility, fn->name) == 0) {
+			facility = fn->code;
+			break;
+		}
+
+	openlog(cf_jobname, LOG_PID, facility);
+	syslog_started = 1;
+}
+
+static void write_syslog(const char *pfx, const char *msg)
+{
+	int prio = LOG_WARNING;
+
+	if (!syslog_started)
+		init_syslog();
+
+	switch (pfx[0]) {
+	case 'F': prio = LOG_CRIT; break;
+	case 'E': prio = LOG_ERR; break;
+	case 'W': prio = LOG_WARNING; break;
+	case 'I': prio = LOG_INFO; break;
+	case 'L': prio = LOG_INFO; break;
+	case 'D': prio = LOG_DEBUG; break;
+	case 'N': prio = LOG_DEBUG; break;
+	}
+
+	syslog(prio, "%s", msg);
+}
 
 void close_logfile(void)
 {
@@ -53,17 +119,22 @@ void close_logfile(void)
 		safe_close(log_fd);
 		log_fd = 0;
 	}
+	close_syslog();
 }
 
 static void write_logfile(const char *buf, int len)
 {
+	int res;
 	if (!log_fd) {
 		int fd = open(cf_logfile, O_CREAT | O_APPEND | O_WRONLY, 0644);
 		if (fd < 0)
 			return;
 		log_fd = fd;
 	}
-	safe_write(log_fd, buf, len);
+	res = safe_write(log_fd, buf, len);
+	if (res < len)
+		/* nothing to do here */
+		len = 0;
 }
 
 static void _log_write(const char *pfx, const char *msg)
@@ -79,6 +150,9 @@ static void _log_write(const char *pfx, const char *msg)
 	if (cf_logfile)
 		write_logfile(buf, len);
 
+	if (cf_syslog)
+		write_syslog(pfx, msg);
+
 	if (!cf_quiet)
 		fprintf(stderr, "%s", buf);
 }
@@ -133,8 +207,7 @@ void log_level(const char *pfx, const char *fmt, ...)
  * Logging about specific PgSocket
  */
 
-void
-slog_level(const char *pfx, const PgSocket *sock, const char *fmt, ...)
+void slog_level(const char *pfx, const PgSocket *sock, const char *fmt, ...)
 {
 	char buf1[1024];
 	char buf2[1024];
@@ -252,8 +325,8 @@ loop:
 
 	if (res < 0) {
 		log_warning("safe_sendmsg(%d, msg[%d,%d], %d) = %s", fd,
-			    msg->msg_iov[0].iov_len,
-			    msg->msg_controllen,
+			    (int)msg->msg_iov[0].iov_len,
+			    (int)msg->msg_controllen,
 			    flags, strerror(errno));
 
 		/* with ancillary data on blocking socket OSX returns
@@ -270,6 +343,49 @@ loop:
 	return res;
 }
 
+static const char *sa2str(const struct sockaddr *sa)
+{
+	static char buf[256];
+	if (sa->sa_family == AF_UNIX) {
+		struct sockaddr_un *un = (struct sockaddr_un *)sa;
+		snprintf(buf, sizeof(buf), "unix:%s", un->sun_path);
+	} else if (sa->sa_family == AF_INET) {
+		struct sockaddr_in *in = (struct sockaddr_in *)sa;
+		snprintf(buf, sizeof(buf), "%s:%d", inet_ntoa(in->sin_addr), ntohs(in->sin_port));
+	} else {
+		snprintf(buf, sizeof(buf), "sa2str: unknown proto");
+	}
+	return buf;
+}
+
+int safe_connect(int fd, const struct sockaddr *sa, socklen_t sa_len)
+{
+	int res;
+loop:
+	res = connect(fd, sa, sa_len);
+	if (res < 0 && errno == EINTR)
+		goto loop;
+	if (res < 0 && (errno != EINPROGRESS || cf_verbose > 2))
+		log_noise("connect(%d, %s) = %s", fd, sa2str(sa), strerror(errno));
+	else if (cf_verbose > 2)
+		log_noise("connect(%d, %s) = %d", fd, sa2str(sa), res);
+	return res;
+}
+
+int safe_accept(int fd, struct sockaddr *sa, socklen_t *sa_len_p)
+{
+	int res;
+loop:
+	res = accept(fd, sa, sa_len_p);
+	if (res < 0 && errno == EINTR)
+		goto loop;
+	if (res < 0)
+		log_noise("safe_accept(%d) = %s", fd, strerror(errno));
+	else if (cf_verbose > 2)
+		log_noise("safe_accept(%d) = %d (%s)", fd, res, sa2str(sa));
+	return res;
+}
+
 /*
  * Load a file into malloc()-ed C string.
  */
@@ -287,8 +403,10 @@ char *load_file(const char *fn)
 	}
 
 	buf = malloc(st.st_size + 1);
-	if (!buf)
+	if (!buf) {
+		log_error("%s: no mem", fn);
 		goto load_error;
+	}
 
 	if ((fd = open(fn, O_RDONLY)) < 0) {
 		log_error("%s: %s", fn, strerror(errno));
@@ -315,7 +433,7 @@ load_error:
  * PostgreSQL MD5 hashing.
  */
 
-static void hash2hex(const uint8 *hash, char *dst)
+static void hash2hex(const uint8_t *hash, char *dst)
 {
 	int i;
 	static const char hextbl [] = "0123456789abcdef";
@@ -326,12 +444,12 @@ static void hash2hex(const uint8 *hash, char *dst)
 	*dst = 0;
 }
 
-bool pg_md5_encrypt(const char *part1,
+void pg_md5_encrypt(const char *part1,
 		    const char *part2, size_t part2len,
 		    char *dest)
 {
 	MD5_CTX ctx;
-	uint8 hash[MD5_DIGEST_LENGTH];
+	uint8_t hash[MD5_DIGEST_LENGTH];
 
 	MD5_Init(&ctx);
 	MD5_Update(&ctx, part1, strlen(part1));
@@ -340,17 +458,14 @@ bool pg_md5_encrypt(const char *part1,
 
 	memcpy(dest, "md5", 3);
 	hash2hex(hash, dest + 3);
-
-	return true;
 }
 
 /* wrapped for getting random bytes */
-bool get_random_bytes(uint8 *dest, int len)
+void get_random_bytes(uint8_t *dest, int len)
 {
 	int i;
 	for (i = 0; i < len; i++)
 		dest[i] = random() & 255;
-	return len;
 }
 
 /*
@@ -585,3 +700,46 @@ void fill_local_addr(PgSocket *sk, int fd, bool is_unix)
 	}
 }
 
+/*
+ * Error handling around evtimer_add() is nasty as the code
+ * may not be called again.  As there is fixed number of timers
+ * in pgbouncer, provider safe_evtimer_add() that stores args of
+ * failed calls in static array and retries later.
+ */
+#define TIMER_BACKUP_SLOTS  10
+
+struct timer_slot {
+	struct event *ev;
+	struct timeval tv;
+};
+static struct timer_slot timer_backup_list[TIMER_BACKUP_SLOTS];
+static int timer_backup_used = 0;
+
+void safe_evtimer_add(struct event *ev, struct timeval *tv)
+{
+	int res;
+	struct timer_slot *ts;
+	
+	res = evtimer_add(ev, tv);
+	if (res >= 0)
+		return;
+
+	if (timer_backup_used >= TIMER_BACKUP_SLOTS)
+		fatal_perror("TIMER_BACKUP_SLOTS full");
+
+	ts = &timer_backup_list[timer_backup_used++];
+	ts->ev = ev;
+	ts->tv = *tv;
+}
+
+void rescue_timers(void)
+{
+	struct timer_slot *ts;
+	while (timer_backup_used) {
+		ts = &timer_backup_list[timer_backup_used - 1];
+		if (evtimer_add(ts->ev, &ts->tv) < 0)
+			break;
+		timer_backup_used--;
+	}
+}
+
