@@ -105,7 +105,7 @@ static int apply_var(PktBuf *pkt, const char *key,
 {
 	char buf[128];
 	char qbuf[128];
-	int len;
+	unsigned len;
 
 	if (strcasecmp(cval, sval) == 0)
 		return 0;
@@ -132,11 +132,11 @@ static int apply_var(PktBuf *pkt, const char *key,
 bool varcache_apply(PgSocket *server, PgSocket *client, bool *changes_p)
 {
 	PktBuf pkt;
-	uint8 buf[1024];
+	uint8_t buf[1024];
 	int changes = 0;
 	const char *cval, *sval;
 	const struct var_lookup *lk;
-	uint8 *debug_sql;
+	uint8_t *debug_sql;
 	bool std_quote = is_std_quote(&server->vars);
 
 	pktbuf_static(&pkt, buf, sizeof(buf));
@@ -169,7 +169,7 @@ void varcache_fill_unset(VarCache *src, PgSocket *dst)
 		srcval = get_value(src, lk);
 		dstval = get_value(&dst->vars, lk);
 		if (!*dstval)
-			strlcpy(dstval, srcval, lk->len);
+			memcpy(dstval, srcval, lk->len);
 	}
 }
 
